'''
By Christopher L. Dancy,
    christopher.dancy@bucknell.edu,
    Assistant Professor @ Bucknell University

Class to simulate Math Model w/ a given penalty distribution &
Function to run Model in MindModeling simulation
'''

from DeckMathModel import *
import numpy
import time
from UtilityFuncs import *

class IGTSim(object):
    ''' class used to run a IGT Simulation w/ a DeckMathModel type object
    '''

    #Number of trials in each block
    BLOCK_SIZE = 20

    def __init__(self, decks, math_model, image_set=None):
        '''
        Arguments:
            decks -- object that contains reward and punishment
            math_model -- The Math Model object (should be an instance of a
             math_model)
        '''
        self.decks = decks
        self.math_model = math_model
        self.image_set = image_set

    def run_sim(self, num_trials, num_runs):
        ''' runs a simulation w/ numTrials, numRuns times
        Arguments:
            num_trials -- Total number of trials per run
                         (a trial is one card selected from a deck)
            num_runs -- Total number of runs per simulation

        Output:
            Dictionary of deck selection proportions for each block
             each key is in the form "b(BLOCKNUMBER)-(DECKNUMBER)"
             s.t., BLOCKNUMBER contained in {0,...,num_trials//BLOCK_SIZE} &
             s.t., DECKNUMBER conatined in {0,...,decks.get_num_decks()}
             for example - b0-0 is block 0, deck 0
        '''
        #Dictionary that holds deck selection proportions for each block
        deck_prop_save = {}
        num_blocks = num_trials//IGTSim.BLOCK_SIZE
        for i in range(num_blocks):
            for j in range(self.decks.get_num_decks()):
                deck_prop_save["b"+str(i+1)+"-"+str(j)] = 0

        #deck_prop_save = Array2D((num_trials//IGTSim.BLOCK_SIZE), self.decks.get_num_decks(), True)
        for i in range(num_runs):
            #Set a seed so we can get reproducible results
            self.math_model.set_model_seed(i+1)

            if (not self.image_set is None):
                #initial image selection from set
                initial_index = self.math_model.get_rand_state().randint(0, len(self.image_set.get_image_set()))
                self.image_set.curr_ind = initial_index

            #initialize values
            total_score = 0;
            deck_exp = [1] * self.decks.get_num_decks()
            n_cards_selected = [0] * self.decks.get_num_decks()

            #Simulate trials
            for j in range(num_trials):
                deck_s = self.math_model.select_deck(self.decks.get_num_decks(), \
                                            n_cards_selected, \
                                            self.decks.get_deck_size(), \
                                            j, deck_exp)

                rew = self.decks.reward[deck_s][(n_cards_selected[deck_s]-1)]
                punish = self.decks.punish[deck_s][(n_cards_selected[deck_s]-1)]

                if (not self.image_set is None):
                    image_val = self.image_set.select_image()
                else:
                    image_val = None

                util = self.math_model.get_util(rew, punish, image_val)
                self.math_model.update_expectancy(deck_s, util, deck_exp)
                deck_prop_save["b"+str((j//IGTSim.BLOCK_SIZE)+1)+"-"+str(deck_s)] += 1

        return (deck_prop_save)




# run
# This function takes two inputs and returns the result of several mathematical operations.
#
# IMPORTANT: The formal arguments here must match the IV names specified during job submission exactly,
# but order does not matter.

def run_ev(update_choice, c, a, weight):
    '''Function to run in MindModeling environment

    Arguments:
        'updateChoice' - should be numbers to represent different
            possible combinations of model choices:
            ((DELta Update equation OR DECay Update equation) AND
            (trial-dependent choice rule OR trial-independent choice rule)
        c - 'Sensitivity' parameter (the higher, the more deterministic)
        a - updating or recency parameter (this is currently alpha in the
             actual mathModel file)
        weight - 'Attention to loss' parameter
    Output:
        Dictionary of deck selection proportions for each block
         each key is in the form "b(BLOCKNUMBER)-(DECKNUMBER)"
         s.t., BLOCKNUMBER contained in {0,...,num_trials//BLOCK_SIZE} &
         s.t., DECKNUMBER conatined in {0,...,decks.get_num_decks()}
         for example - b0-0 is block 0, deck 0
    '''

    #Convert IVs to required types
    update_choice = int(update_choice)
    c = float(c)
    a = float(a)
    weight = float(weight)

    if (update_choice == 0):
        update_fun = DeckMathModel._delta_update_fun
        choice_fun = DeckMathModel._trial_dep
    elif (update_choice == 1):
        update_fun = DeckMathModel._delta_update_fun
        choice_fun = DeckMathModel._trial_indep
    elif (update_choice == 2):
        update_fun = DeckMathModel._decay_update_fun
        choice_fun = DeckMathModel._trial_dep
    else:
        update_fun = DeckMathModel._decay_update_fun
        choice_fun = DeckMathModel._trial_indep


    my_sim = IGTSim(Decks(), EVMathModel("MindModelingRun", c, a, update_fun, \
                                        choice_fun, weight))
    num_runs = 100

    num_trials = 100

    dvs = my_sim.run_sim(num_trials, num_runs)
    fixed_dvs = {}

    # Convert our DVs to what MindModeling expects
    for key in dvs.keys():
        #Normalize so we get a proportion for each deck w/ that block
        dvs[key] = dvs[key] / float((num_runs * IGTSim.BLOCK_SIZE))
        if (str.find(key, "-0") != -1):
            fixed_dvs[str.replace(key, "-0", "A")] = dvs[key]
        elif (str.find(key, "-1") != -1):
            fixed_dvs[str.replace(key, "-1", "B")] = dvs[key]
        elif (str.find(key, "-2") != -1):
            fixed_dvs[str.replace(key, "-2", "C")] = dvs[key]
        elif (str.find(key, "-3") != -1):
            fixed_dvs[str.replace(key, "-3", "D")] = dvs[key]

    return (fixed_dvs)

def run_ev_affect(update_choice, c, a, weight, affect_weight="0", group="2",\
                    sex="1"):
    '''Function to run in MindModeling environment

    Arguments:
        'updateChoice' - should be numbers to represent different
            possible combinations of model choices:
            ((DELta Update equation OR DECay Update equation) AND
            (trial-dependent choice rule OR trial-independent choice rule)
        c - 'Sensitivity' parameter (the higher, the more deterministic)
        a - updating or recency parameter (this is currently alpha in the
             actual mathModel file)
        weight - 'Attention to loss' parameter
        affect_weight - Weight of affective value (images in our case)
    Output:
        Dictionary of deck selection proportions for each block
         each key is in the form "b(BLOCKNUMBER)-(DECKNUMBER)"
         s.t., BLOCKNUMBER contained in {0,...,num_trials//BLOCK_SIZE} &
         s.t., DECKNUMBER conatined in {0,...,decks.get_num_decks()}
         for example - b0-0 is block 0, deck 0
    '''

    #Set ImageSet Function based on group
    if (group == "1"):
        image_set_fun = ImageSet._IAPS_Negative_Image_Set
    elif (group == "3"):
        image_set_fun = ImageSet._IAPS_Positive_Image_Set
    else:
        image_set_fun = ImageSet._IAPS_Neutral_Image_Set

    #Set sex based on argument
    if (sex == "1"):
        sex = "Male"
    else:
        sex == "Female"

    #Convert IVs to required types
    update_choice = int(update_choice)
    c = float(c)
    a = float(a)
    weight = float(weight)
    affect_weight=float(affect_weight)

    if (update_choice == 0):
        update_fun = DeckMathModel._delta_update_fun
        choice_fun = DeckMathModel._trial_dep
    elif (update_choice == 1):
        update_fun = DeckMathModel._delta_update_fun
        choice_fun = DeckMathModel._trial_indep
    elif (update_choice == 2):
        update_fun = DeckMathModel._decay_update_fun
        choice_fun = DeckMathModel._trial_dep
    else:
        update_fun = DeckMathModel._decay_update_fun
        choice_fun = DeckMathModel._trial_indep


    #print("Run w/ Params - " + str(update_choice) + ", " + str(c) + ", " \
    #      + str(a) + ", " + str(weight))


    my_sim = IGTSim(Decks(), EVMathModel("MindModelingRun", c, a, update_fun, \
                                        choice_fun, weight, affect_weight), \
                                        ImageSet(0, 5, image_set_fun, sex))
    num_runs = 100
    num_trials = 100

    dvs = my_sim.run_sim(num_trials, num_runs)
    fixed_dvs = {}

    # Convert our DVs to what MindModeling expects
    for key in dvs.keys():
        #Normalize so we get a proportion for each deck w/ that block
        dvs[key] = dvs[key] / float((num_runs * IGTSim.BLOCK_SIZE))
        if (str.find(key, "-0") != -1):
            fixed_dvs[str.replace(key, "-0", "A")] = dvs[key]
        elif (str.find(key, "-1") != -1):
            fixed_dvs[str.replace(key, "-1", "B")] = dvs[key]
        elif (str.find(key, "-2") != -1):
            fixed_dvs[str.replace(key, "-2", "C")] = dvs[key]
        elif (str.find(key, "-3") != -1):
            fixed_dvs[str.replace(key, "-3", "D")] = dvs[key]

    return (fixed_dvs)

def run_ev_affect_all(c, a, weight, affect_weight="0"):
    '''Function to run in MindModeling environment

    Arguments:
        'updateChoice' - should be numbers to represent different
            possible combinations of model choices:
            ((DELta Update equation OR DECay Update equation) AND
            (trial-dependent choice rule OR trial-independent choice rule)
            THIS IS NOW DEPRECATED (Code uses loop)
        c - 'Sensitivity' parameter (the higher, the more deterministic)
        a - updating or recency parameter (this is currently alpha in the
             actual mathModel file)
        weight - 'Attention to loss' parameter
        affect_weight - Weight of affective value (images in our case)
    Output:
        Dictionary of deck selection proportions for each block
         each key is in the form "UC(CHOICE)_G(GROUP)_S(SEX)_b(BLOCKNUMBER)-(DECKNUMBER)"
         s.t., BLOCKNUMBER contained in {0,...,num_trials//BLOCK_SIZE} &
         s.t., DECKNUMBER conatined in {0,...,decks.get_num_decks()}
         for example - b0-0 is block 0, deck 0
    '''
    fixed_dvs = {}
    for i in range(4):
        for j in range(3):
            for k in range(2):
                #Set ImageSet Function based on group
                if (j == 0):
                    image_set_fun = ImageSet._IAPS_Negative_Image_Set
                elif (j == 2):
                    image_set_fun = ImageSet._IAPS_Positive_Image_Set
                else:
                    image_set_fun = ImageSet._IAPS_Neutral_Image_Set

                #Set sex based on argument
                if (k == 0):
                    sex = "Male"
                else:
                    sex == "Female"

                #Convert IVs to required types
                c = float(c)
                a = float(a)
                weight = float(weight)
                affect_weight=float(affect_weight)

                if (i == 0):
                    update_fun = DeckMathModel._delta_update_fun
                    choice_fun = DeckMathModel._trial_dep
                elif (i == 1):
                    update_fun = DeckMathModel._delta_update_fun
                    choice_fun = DeckMathModel._trial_indep
                elif (i == 2):
                    update_fun = DeckMathModel._decay_update_fun
                    choice_fun = DeckMathModel._trial_dep
                else:
                    update_fun = DeckMathModel._decay_update_fun
                    choice_fun = DeckMathModel._trial_indep


                my_sim = IGTSim(Decks(), EVMathModel("MindModelingRun", c, a, update_fun, \
                                                    choice_fun, weight, affect_weight), \
                                                    ImageSet(0, 5, image_set_fun, sex))
                num_runs = 100
                num_trials = 100

                dvs = my_sim.run_sim(num_trials, num_runs)

                # Convert our DVs to what MindModeling expects
                for key in dvs.keys():
                    dvs = my_sim.run_sim(num_trials, num_runs)
                    #Normalize so we get a proportion for each deck w/ that block
                    dvs[key] = dvs[key] / float((num_runs * IGTSim.BLOCK_SIZE))
                    if (str.find(key, "-0") != -1):
                        fixed_dvs["UC" + str(i+1) + "_G" + str(j+1) + "_S" + str(k+1) \
                         + "_" + str.replace(key, "-0", "A")] = dvs[key]
                    elif (str.find(key, "-1") != -1):
                        fixed_dvs["UC" + str(i+1) + "_G" + str(j+1) + "_S" + str(k+1) \
                         + "_" + str.replace(key, "-1", "B")] = dvs[key]
                    elif (str.find(key, "-2") != -1):
                        fixed_dvs["UC" + str(i+1) + "_G" + str(j+1) + "_S" + str(k+1) \
                         + "_" + str.replace(key, "-2", "C")] = dvs[key]
                    elif (str.find(key, "-3") != -1):
                        fixed_dvs["UC" + str(i+1) + "_G" + str(j+1) + "_S" + str(k+1) \
                         + "_" + str.replace(key, "-3", "D")] = dvs[key]
    return (fixed_dvs)

def run_pv_affect_all(c, a, alpha, loss_aver, affect_weight=0):
    '''Function to run in MindModeling environment

    Arguments:
        'updateChoice' - should be numbers to represent different
            possible combinations of model choices:
            ((DELta Update equation OR DECay Update equation) AND
            (trial-dependent choice rule OR trial-independent choice rule)
            THIS IS NOW DEPRECATED (Code uses loop)
        c - 'Sensitivity' parameter (the higher, the more deterministic)
        a - updating or recency parameter (this is currently alpha in the
             actual mathModel file)
        alpha - controls shape of utility function
        loss_aver - loss aversion parameter
        affect_weight - Weight of affective value (images in our case)
    Output:
        Dictionary of deck selection proportions for each block
         each key is in the form "b(BLOCKNUMBER)-(DECKNUMBER)"
         s.t., BLOCKNUMBER contained in {0,...,num_trials//BLOCK_SIZE} &
         s.t., DECKNUMBER conatined in {0,...,decks.get_num_decks()}
         for example - b0-0 is block 0, deck 0
    '''

    fixed_dvs = {}
    for i in range(4):
        for j in range(3):
            for k in range(2):
                #Set ImageSet Function based on group
                if (j == 0):
                    image_set_fun = ImageSet._IAPS_Negative_Image_Set
                elif (j == 2):
                    image_set_fun = ImageSet._IAPS_Positive_Image_Set
                else:
                    image_set_fun = ImageSet._IAPS_Neutral_Image_Set

                #Set sex based on argument
                if (k == 0):
                    sex = "Male"
                else:
                    sex == "Female"

                #Convert IVs to required types
                c = float(c)
                a = float(a)
                alpha = float(alpha)
                loss_aver = float(loss_aver)
                affect_weight=float(affect_weight)

                if (i == 0):
                    update_fun = DeckMathModel._delta_update_fun
                    choice_fun = DeckMathModel._trial_dep
                elif (i == 1):
                    update_fun = DeckMathModel._delta_update_fun
                    choice_fun = DeckMathModel._trial_indep
                elif (i == 2):
                    update_fun = DeckMathModel._decay_update_fun
                    choice_fun = DeckMathModel._trial_dep
                else:
                    update_fun = DeckMathModel._decay_update_fun
                    choice_fun = DeckMathModel._trial_indep


                my_sim = IGTSim(Decks(), PVMathModel("MindModelingRun", c, a, update_fun, \
                                                    choice_fun, alpha, loss_aver, affect_weight), \
                                                    ImageSet(0, 5, image_set_fun, sex))
                num_runs = 100
                num_trials = 100

                dvs = my_sim.run_sim(num_trials, num_runs)

                # Convert our DVs to what MindModeling expects
                for key in dvs.keys():
                    dvs = my_sim.run_sim(num_trials, num_runs)
                    #Normalize so we get a proportion for each deck w/ that block
                    dvs[key] = dvs[key] / float((num_runs * IGTSim.BLOCK_SIZE))
                    if (str.find(key, "-0") != -1):
                        fixed_dvs["UC" + str(i+1) + "_G" + str(j+1) + "_S" + str(k+1) \
                         + "_" + str.replace(key, "-0", "A")] = dvs[key]
                    elif (str.find(key, "-1") != -1):
                        fixed_dvs["UC" + str(i+1) + "_G" + str(j+1) + "_S" + str(k+1) \
                         + "_" + str.replace(key, "-1", "B")] = dvs[key]
                    elif (str.find(key, "-2") != -1):
                        fixed_dvs["UC" + str(i+1) + "_G" + str(j+1) + "_S" + str(k+1) \
                         + "_" + str.replace(key, "-2", "C")] = dvs[key]
                    elif (str.find(key, "-3") != -1):
                        fixed_dvs["UC" + str(i+1) + "_G" + str(j+1) + "_S" + str(k+1) \
                         + "_" + str.replace(key, "-3", "D")] = dvs[key]

    return (fixed_dvs)


def run_pv(update_choice, c, a, alpha, loss_aver):
    '''Function to run in MindModeling environment

    Arguments:
        'updateChoice' - should be numbers to represent different
            possible combinations of model choices:
            ((DELta Update equation OR DECay Update equation) AND
            (trial-dependent choice rule OR trial-independent choice rule)
        c - 'Sensitivity' parameter (the higher, the more deterministic)
        a - updating or recency parameter (this is currently alpha in the
             actual mathModel file)
        alpha - controls shape of utility function
        loss_aver - loss aversion parameter
    Output:
        Dictionary of deck selection proportions for each block
         each key is in the form "b(BLOCKNUMBER)-(DECKNUMBER)"
         s.t., BLOCKNUMBER contained in {0,...,num_trials//BLOCK_SIZE} &
         s.t., DECKNUMBER conatined in {0,...,decks.get_num_decks()}
         for example - b0-0 is block 0, deck 0
    '''

    #Convert IVs to required types
    update_choice = int(update_choice)
    c = float(c)
    a = float(a)
    alpha = float(alpha)
    loss_aver = float(loss_aver)

    if (update_choice == 0):
        update_fun = DeckMathModel._delta_update_fun
        choice_fun = DeckMathModel._trial_dep
    elif (update_choice == 1):
        update_fun = DeckMathModel._delta_update_fun
        choice_fun = DeckMathModel._trial_indep
    elif (update_choice == 2):
        update_fun = DeckMathModel._decay_update_fun
        choice_fun = DeckMathModel._trial_dep
    else:
        update_fun = DeckMathModel._decay_update_fun
        choice_fun = DeckMathModel._trial_indep


    my_sim = IGTSim(Decks(), PVMathModel("MindModelingRun", c, a, update_fun, \
                                        choice_fun, alpha, loss_aver))
    num_runs = 100
    num_trials = 100

    dvs = my_sim.run_sim(num_trials, num_runs)
    fixed_dvs = {}

    # Convert our DVs to what MindModeling expects
    for key in dvs.keys():
        #Normalize so we get a proportion for each deck w/ that block
        dvs[key] = dvs[key] / float((num_runs * IGTSim.BLOCK_SIZE))
        if (str.find(key, "-0") != -1):
            fixed_dvs[str.replace(key, "-0", "A")] = dvs[key]
        elif (str.find(key, "-1") != -1):
            fixed_dvs[str.replace(key, "-1", "B")] = dvs[key]
        elif (str.find(key, "-2") != -1):
            fixed_dvs[str.replace(key, "-2", "C")] = dvs[key]
        elif (str.find(key, "-3") != -1):
            fixed_dvs[str.replace(key, "-3", "D")] = dvs[key]


    return (fixed_dvs)

def run_pv_affect(update_choice, c, a, alpha, loss_aver, affect_weight=0, \
                    group="2", sex="1"):
    '''Function to run in MindModeling environment

    Arguments:
        'updateChoice' - should be numbers to represent different
            possible combinations of model choices:
            ((DELta Update equation OR DECay Update equation) AND
            (trial-dependent choice rule OR trial-independent choice rule)
        c - 'Sensitivity' parameter (the higher, the more deterministic)
        a - updating or recency parameter (this is currently alpha in the
             actual mathModel file)
        alpha - controls shape of utility function
        loss_aver - loss aversion parameter
        affect_weight - Weight of affective value (images in our case)
    Output:
        Dictionary of deck selection proportions for each block
         each key is in the form "b(BLOCKNUMBER)-(DECKNUMBER)"
         s.t., BLOCKNUMBER contained in {0,...,num_trials//BLOCK_SIZE} &
         s.t., DECKNUMBER conatined in {0,...,decks.get_num_decks()}
         for example - b0-0 is block 0, deck 0
    '''

    #Set ImageSet Function based on group
    if (group == "1"):
        image_set_fun = ImageSet._IAPS_Negative_Image_Set
    elif (group == "3"):
        image_set_fun = ImageSet._IAPS_Positive_Image_Set
    else:
        image_set_fun = ImageSet._IAPS_Neutral_Image_Set

    #Set sex based on argument
    if (sex == "1"):
        sex = "Male"
    else:
        sex == "Female"

    #Convert IVs to required types
    update_choice = int(update_choice)
    c = float(c)
    a = float(a)
    alpha = float(alpha)
    loss_aver = float(loss_aver)
    affect_weight=float(affect_weight)

    if (update_choice == 0):
        update_fun = DeckMathModel._delta_update_fun
        choice_fun = DeckMathModel._trial_dep
    elif (update_choice == 1):
        update_fun = DeckMathModel._delta_update_fun
        choice_fun = DeckMathModel._trial_indep
    elif (update_choice == 2):
        update_fun = DeckMathModel._decay_update_fun
        choice_fun = DeckMathModel._trial_dep
    else:
        update_fun = DeckMathModel._decay_update_fun
        choice_fun = DeckMathModel._trial_indep


    my_sim = IGTSim(Decks(), PVMathModel("MindModelingRun", c, a, update_fun, \
                                        choice_fun, alpha, loss_aver, affect_weight), \
                                        ImageSet(0, 5, image_set_fun, sex))
    num_runs = 2
    num_trials = 100

    dvs = my_sim.run_sim(num_trials, num_runs)
    fixed_dvs = {}

    # Convert our DVs to what MindModeling expects
    for key in dvs.keys():
        #Normalize so we get a proportion for each deck w/ that block
        dvs[key] = dvs[key] / float((num_runs * IGTSim.BLOCK_SIZE))
        if (str.find(key, "-0") != -1):
            fixed_dvs[str.replace(key, "-0", "A")] = dvs[key]
        elif (str.find(key, "-1") != -1):
            fixed_dvs[str.replace(key, "-1", "B")] = dvs[key]
        elif (str.find(key, "-2") != -1):
            fixed_dvs[str.replace(key, "-2", "C")] = dvs[key]
        elif (str.find(key, "-3") != -1):
            fixed_dvs[str.replace(key, "-3", "D")] = dvs[key]

    return (fixed_dvs)


def frange(start, stop, step):
    i = start
    while i < stop:
        yield i
        i += step

def test_ev_fun():
    for i in range(0, 4, 1):
        for j in frange(-10, 10, 0.5):
            for k in frange(0, 1, 0.05):
                for l in frange(0, 1, 0.05):
                    run_ev(i, j, k, l)

def test_ev_affect_fun():
    for i in range(0, 4, 1):
        for j in frange(-10, 10, 0.5):
            for k in frange(0, 1, 0.5):
                for l in frange(0, 1, 0.5):
                    for a in frange(-10,10,0.5):
                        run_ev_affect(i, j, k, l, a)

def test_pv_fun():
    for i in range(0, 4, 1):
        for j in frange(-10, 10, 0.5):
            for k in frange(0, 1, 0.1):
                for l in frange(0, 1, 0.05):
                    for m in frange(0, 10, 0.5):
                        run_pv(i, j, k, l, m)

def test_pv_affect_fun():
    for i in range(0, 4, 1):
        for j in frange(-10, 10, 0.5):
            for k in frange(0, 1, 0.5):
                for l in frange(0, 1, 0.5):
                    for m in frange(0, 10, 0.5):
                        for a in frange(-10,10,1):
                            run_pv_affect(i, j, k, l, m, a)

#test_time = time.time()
#run_ev_affect_all(1.5,.55,.25,-2.5)
#test_pv_affect_fun()
#run_ev(0,-10,0,0)
#print(str(time.time() - test_time))

