'''
By Christopher L. Dancy][
    christopher.dancy@bucknell.edu,
    Assistant Professor @ Bucknell University

Class to create a model that simulates deck selection in a version of the
 Iowa Gambling Task (IGT)

A utility function (Decks) is also included here that can be
 changed/extended by anyone wishing to use a version of the IGT w/ a different
 distribution (or a different )
'''

import abc
import numpy
import math
import random

class DeckMathModel(object):
    ''' Parent class to build a Model that simulates deck selection from a gambling task
     (that is, or is similar to, the IGT)
    '''
    #So we can use abstract methods
    __metaclass__  = abc.ABCMeta
    def __init__(self, id="GenericMathModel", c=0, a=0, update_fun=None, \
                choice_fun=None):
        ''' All of our models have an ID, a consistency param, and a learning rate param

        Arguments:
            c   --  consistency parameter
            a   --  learning/decay rate
            updateFun --  update function used to update the expectancy
            choice_fun -- function used to find parameter important for deck
                            choice probabilities
        '''
        self.id = id
        self.c = c
        self.a = a
        self._rand_state = numpy.random.RandomState()
        self.update_fun = update_fun
        if (self.update_fun is None):
            self.update_fun = DeckMathModel._delta_update_fun
        self.choice_fun = choice_fun
        if (self.choice_fun is None):
            self.choice_fun = DeckMathModel._trial_dep

    def select_deck(self, num_decks, n_cards_selected, max_cards, curr_trial, \
                    expectancy):
        '''chooses a deck based on current state and returns selection
            (also updates the current_trial)
        Ouput:
            python list w/ integer that represents the deck chosen by the model
             and current trial
        '''
        #We use a pseudorandom number as our token, to signify when we have
        # decks that should have a probability of 0 due to the maximum number of
        # cards being selected from that deck
        deck_prob_token = random.random()

        #Calculate Phi for probability
        phi = self.choice_fun(self.c, curr_trial)
        deck_probs = [0] * num_decks
        deck_select = -1

        for i in range(num_decks):
            if (n_cards_selected[i] < max_cards):
                deck_probs[i] = phi * expectancy[i]
            else:
                deck_probs[i] = deck_prob_token

        #non-zero deck probabilities
        nzdp = [x for x in deck_probs if x != deck_prob_token]

        temp_deck_probs = deck_probs[:]

        for i in range(num_decks):
            if( deck_probs[i] != deck_prob_token ):
                deck_probs[i] = math.exp(deck_probs[i] - max(nzdp)) / \
                                math.fsum([math.exp(x - max(nzdp)) for x in nzdp])
            else:
                deck_probs[i] = 0

        while deck_select == -1 or n_cards_selected[deck_select] >= max_cards:
            deck_select = list(self._rand_state.multinomial(1, deck_probs))
            deck_select = deck_select.index(1)
        n_cards_selected[deck_select] += 1
        return (deck_select)

    def get_rand_state(self):
        return (self._rand_state)

    def set_model_seed(self, seed):
        self._rand_state.seed(seed)

    @abc.abstractmethod
    def get_util(self, rew, punish, image_val):
        '''Method to get utility based on the type of model (e.g., EV)

        Arguments:
            rew   --  reward
            punish -- punishment
            image_val -- Given if affective image is given w/ reward/punishment
        Output:
            integer that represents the new utility
        '''
        pass


    def update_expectancy(self, deck_s, deck_util, deck_exp):
        ''' Updates expectancy based on current internal state

        Arguments:
            deck_s  -- integer representing the deck selected
            deck_util   -- Utility for selected deck
            deck_exp    --  Expectancy array
        '''

        #We use the internal update expectancy function to do the actual math
        self.update_fun(deck_s, deck_util, deck_exp, self.a)

    # -- Utility (static) methods

    @staticmethod
    def _trial_dep(c,trial):
        ''' Function to calculate parameter for trial-dependent deck probability
                calculation
        Arguments:
            c   --  consistency parameter
            trial   --  current trial
        '''
        phi = 0
        #Can't raise 0 to any number but 0 (in which case, the result is 1)
        if (trial / 10.0 != 0):
            phi = (trial/10.0) ** c
        elif (c == 0):
            phi = 1
        return (phi)

    @staticmethod
    def _trial_indep(c,trial=0):
        ''' Function to calculate parameter for trial-independent deck
                probability calculation
        Arguments:
            c   --  consistency parameter
            trial   --  current trial (not used here, but kept for consistency)
        '''
        return (3 ** (c - 1.0))

    @staticmethod
    def _delta_update_fun(deck_s, util, expectancy, alpha):
        ''' Function to update expectancy according to delta function

        Arguments:
            deck_s -- integer representing the deck selected
            util -- Utility for selected deck
            expectancy  --  Python list that represents the current expectancies
                                for each deck
            alpha -- learning rate parameter
        '''
        expectancy[deck_s] += alpha * (util - expectancy[deck_s])

    @staticmethod
    def _decay_update_fun(deck_s, util, expectancy, alpha):
        ''' Function to update expectancy according to delta function

        Arguments:
            deck_s -- integer representing the deck selected
            util -- Utility for selected deck
            expectancy  --  Python list that represents the current expectancies
                                for each deck
            alpha -- decay rate parameter
        '''
        #Apply decay to all values in x
        expectancy[:] = [x*alpha for x in expectancy]
        #Add utility to deck selected
        expectancy[deck_s] += util

class EVMathModel(DeckMathModel):
    ''' Implements expectancy valence model
    '''

    def __init__(self, id="GenericEVModel", c=0.32, a=0.34, update_fun=None, \
                choice_fun=None, weight=0.35, affect_weight=0):
                super(EVMathModel, self).__init__(id,c,a,update_fun,choice_fun)
                self.weight = weight
                #Controls how much images affect reward/loss
                self.affect_weight = affect_weight
                if (affect_weight == 0 or affect_weight is None):
                    self.util_fun = self.norm_util
                else:
                    self.util_fun = self.affective_util

    def get_util(self, rew, punish, image_val=0):
        '''Method to get utility based on the util_fun attribute

        Arguments:
            rew   --  reward
            punish -- punishment
            image_val -- Given if affective image is given w/ reward/punishment
        Output:
            integer that represents the new utility
        '''
        return (self.util_fun(rew,punish,image_val))

    def norm_util(self, rew, punish,image_val):
        '''Method to get utility based on the type of model (e.g., EV)

        Arguments:
            rew   --  reward
            punish -- punishment
            image_val -- Kept for static method, not used here
        Output:
            integer that represents the new utility
        '''
        return (((1 - self.weight) * rew) - (self.weight * punish))


    def affective_util(self, rew, punish, image_val):
        '''Method to get utility based on the type of model (e.g., EV)

        Arguments:
            rew   --  reward
            punish -- punishment
            image_val -- list of arousal, valence, dominance values for image
        Output:
            integer that represents the new utility
        '''
        weight_rew = ((1 - self.weight) * rew)

        weight_pun = (self.weight * punish)

        if (not image_val is None):
            #Get values from normal distribution based on mean and std
            val = self._rand_state.normal(image_val[0], image_val[1])
            aro = self._rand_state.normal(image_val[2], image_val[3])
            dom = self._rand_state.normal(image_val[4], image_val[5])
            #affect_weight * (arousal * (valence+dominance - 10) / 90)
            affect_rew = self.affect_weight * aro * (val + dom - 10) / 90
            #affect_weight * (arousal * (10 - valence - dominance) / 90)
            affect_pun = self.affect_weight * aro * (10 - val - dom) / 90
        else:
            affect_rew = 0
            affect_pun = 0

        return (weight_rew + affect_rew - weight_pun - affect_pun)

class PVMathModel(DeckMathModel):
    ''' Implements prospective valence model
    '''

    def __init__(self, id="GenericEVModel", c=0, a=0, update_fun=None, \
                choice_fun=None, alpha=0, l=0, affect_weight=0):
                super(PVMathModel, self).__init__(id,c,a,update_fun,choice_fun)
                self.alpha = alpha
                self.l = l
                self.affect_weight = affect_weight
                #Use the normal utility if our affect_weight is 0 or None
                if (affect_weight == 0 or affect_weight is None):
                    self.util_fun = self.norm_util
                else:
                    self.util_fun = self.affective_util

    def get_util(self, rew, punish, image_val):
        '''Method to get utility based on the type of model (e.g., EV)

        Arguments:
            rew   --  reward
            punish -- punishment
            image_val -- Given if affective image is given w/ reward/punishment
        Output:
            integer that represents the new utility
        '''
        return (self.util_fun(rew, punish, image_val))

    def norm_util(self, rew, punish, image_val=None):
        '''Method to get utility based on the type of model (e.g., EV)

        Arguments:
            rew   --  reward
            punish -- punishment
            image_val -- Kept for abstract method, not used here
        Output:
            integer that represents the new utility
        '''
        net_gain = rew - punish
        if (net_gain >= 0):
            return (math.pow(net_gain, self.alpha))
        else:
            return (self.l * math.pow(abs(net_gain), self.alpha))

    def affective_util(self, rew, punish, image_val=None):
        '''Method to get utility based on the type of model (e.g., EV)

        Arguments:
            rew   --  reward
            punish -- punishment
            image_val -- list of arousal, valence, dominance values for image
        Output:
            integer that represents the new utility
        '''
        net_gain = rew - punish
        if (not image_val is None):
            #Get values from normal distribution based on mean and std
            val = self._rand_state.normal(image_val[0], image_val[1])
            aro = self._rand_state.normal(image_val[2], image_val[3])
            dom = self._rand_state.normal(image_val[4], image_val[5])
            #affect_weight * (arousal * (valence+dominance - 10) / 90)
            affect_rew = self.affect_weight * aro * (val + dom - 10) / 90
            #affect_weight * (arousal * (10 - valence - dominance) / 90)
            affect_pun = self.affect_weight * aro * (10 - val - dom) / 90
        else:
            affect_rew = 0
            affect_pun = 0

        if (net_gain >= 0):
            return (math.pow(net_gain, self.alpha) \
                    + affect_rew - affect_pun)
        else:
            return (self.l * math.pow(abs(net_gain), self.alpha) \
                    + affect_rew - affect_pun)

class ImageSet(object):
    '''Creates a set of images, each of which have several images that have
        attached values (e.g., IAPS)
    '''
    def __init__(self, starting_image_ind=0, image_set_size=5, \
                set_image_fun=None, target_sex="Male"):
        #Set the initial current image to be selected next
        self.curr_ind = starting_image_ind
        self.image_set = [[0] * 6 for x in range(image_set_size)]
        #Our default image set if the IAPS neutral set
        if (set_image_fun == None):
            self.set_image_fun = ImageSet._IAPS_Neutral_Image_Set_Male
        else:
            self.set_image_fun = set_image_fun
        self.set_image_fun(self.image_set, target_sex)

    def get_image_set(self):
        return (self.image_set)

    def select_image(self):
        '''Returns values of current image and increments index
        '''
        #Just to shorten the lines a bit
        i_s = self.image_set
        c_i = self.curr_ind
        #increment our current index
        self.curr_ind = (self.curr_ind + 1) % len(self.image_set)

        return ([i_s[c_i][0], i_s[c_i][1], i_s[c_i][2], i_s[c_i][3], \
                i_s[c_i][4], i_s[c_i][5]])

    @staticmethod
    def _IAPS_Negative_Image_Set(image_set, sex):
        '''Image set values for 1050, 1202, 1220, 1304, 1525 (Male) &
            1050, 1120, 1201, 1202, 1525 (Female)
            Gives mean and std for each of arousal, valence, and dominance
        '''
        if (len(image_set[0]) != 6):
            print("Only prepared for image_sets with 6 images at the moment....")
            return

        if (sex == "Male"):
            image_set[0][0] = 3.9
            image_set[0][1] = 2.28
            image_set[0][2] = 6.84
            image_set[0][3] = 1.55
            image_set[0][4] = 3.14
            image_set[0][5] = 1.78
            image_set[1][0] = 4.03
            image_set[1][1] = 1.81
            image_set[1][2] = 6.2
            image_set[1][3] = 1.45
            image_set[1][4] = 4.8
            image_set[1][5] = 2.32
            image_set[2][0] = 3.88
            image_set[2][1] = 1.76
            image_set[2][2] = 5.4
            image_set[2][3] = 2.5
            image_set[2][4] = 5.02
            image_set[2][5] = 2.6
            image_set[3][0] = 3.89
            image_set[3][1] = 1.6
            image_set[3][2] = 6.39
            image_set[3][3] = 1.9
            image_set[3][4] = 3.18
            image_set[3][5] = 1.47
            image_set[4][0] = 3.55
            image_set[4][1] = 1.59
            image_set[4][2] = 6.14
            image_set[4][3] = 2.31
            image_set[4][4] = 3.37
            image_set[4][5] = 2.24
        else:
            image_set[0][0] = 3.02
            image_set[0][1] = 1.93
            image_set[0][2] = 6.9
            image_set[0][3] = 1.82
            image_set[0][4] = 3.02
            image_set[0][5] = 2.08
            image_set[1][0] = 3.03
            image_set[1][1] = 1.74
            image_set[1][2] = 7.2
            image_set[1][3] = 1.86
            image_set[1][4] = 3.22
            image_set[1][5] = 2.42
            image_set[2][0] = 2.93
            image_set[2][1] = 1.81
            image_set[2][2] = 6.87
            image_set[2][3] = 2.09
            image_set[2][4] = 3.82
            image_set[2][5] = 2.26
            image_set[3][0] = 2.98
            image_set[3][1] = 1.65
            image_set[3][2] = 5.8
            image_set[3][3] = 2.47
            image_set[3][4] = 3.92
            image_set[3][5] = 2.34
            image_set[4][0] = 2.67
            image_set[4][1] = 1.74
            image_set[4][2] = 6.886
            image_set[4][3] = 2.16
            image_set[4][4] = 2.55
            image_set[4][5] = 1.99

    @staticmethod
    def _IAPS_Neutral_Image_Set(image_set, sex):
        '''Image set values for 1670, 7006, 7010, 7080, 7175 (Male) &
            1670, 7004, 7010, 7012, 7175 (Female)
        '''
        if (len(image_set[0]) != 6):
            print("Only prepared for image_sets with 6 images at the moment....")
            return

        if (sex == "Male"):
            image_set[0][0] = 6.51
            image_set[0][1] = 1.76
            image_set[0][2] = 2.95
            image_set[0][3] = 1.67
            image_set[0][4] = 5.9
            image_set[0][5] = 1.87
            image_set[1][0] = 4.65
            image_set[1][1] = 1.10
            image_set[1][2] = 2.08
            image_set[1][3] = 1.58
            image_set[1][4] = 5.83
            image_set[1][5] = 2.03
            image_set[2][0] = 4.95
            image_set[2][1] = 1.43
            image_set[2][2] = 1.55
            image_set[2][3] = 1.36
            image_set[2][4] = 7.25
            image_set[2][5] = 2.31
            image_set[3][0] = 5.43
            image_set[3][1] = 1.26
            image_set[3][2] = 1.98
            image_set[3][3] = 1.63
            image_set[3][4] = 7.47
            image_set[3][5] = 2.10
            image_set[4][0] = 4.78
            image_set[4][1] = 1.18
            image_set[4][2] = 1.55
            image_set[4][3] = 0.96
            image_set[4][4] = 6.51
            image_set[4][5] = 2.08
        else:
            image_set[0][0] = 5.88
            image_set[0][1] = 1.84
            image_set[0][2] = 3.52
            image_set[0][3] = 2.05
            image_set[0][4] = 5.40
            image_set[0][5] = 1.74
            image_set[1][0] = 5.14
            image_set[1][1] = 0.59
            image_set[1][2] = 1.94
            image_set[1][3] = 1.60
            image_set[1][4] = 6.65
            image_set[1][5] = 2.10
            image_set[2][0] = 4.92
            image_set[2][1] = 0.48
            image_set[2][2] = 1.97
            image_set[2][3] = 1.58
            image_set[2][4] = 6.13
            image_set[2][5] = 2.29
            image_set[3][0] = 4.97
            image_set[3][1] = 0.87
            image_set[3][2] = 2.88
            image_set[3][3] = 1.90
            image_set[3][4] = 6.17
            image_set[3][5] = 1.68
            image_set[4][0] = 4.95
            image_set[4][1] = 0.80
            image_set[4][2] = 1.87
            image_set[4][3] = 1.48
            image_set[4][4] = 6.44
            image_set[4][5] = 2.01

    @staticmethod
    def _IAPS_Positive_Image_Set(image_set, sex):
        '''Image set values for 4180, 4210, 4232, 4664, 8501 (Male) &
            4505, 4525, 4660, 8001, 8501 (Female)
        '''
        if (len(image_set[0]) != 6):
            print("Only prepared for image_sets with 6 images at the moment....")
            return

        if (sex == "Male"):
            image_set[0][0] = 8.21
            image_set[0][1] = 1.34
            image_set[0][2] = 7.43
            image_set[0][3] = 1.97
            image_set[0][4] = 6.41
            image_set[0][5] = 2.54
            image_set[1][0] = 8.25
            image_set[1][1] = 1.30
            image_set[1][2] = 7.80
            image_set[1][3] = 1.90
            image_set[1][4] = 6.97
            image_set[1][5] = 2.42
            image_set[2][0] = 7.88
            image_set[2][1] = 1.10
            image_set[2][2] = 7.52
            image_set[2][3] = 1.51
            image_set[2][4] = 6.81
            image_set[2][5] = 2.02
            image_set[3][0] = 7.99
            image_set[3][1] = 1.25
            image_set[3][2] = 7.72
            image_set[3][3] = 1.45
            image_set[3][4] = 6.54
            image_set[3][5] = 1.93
            image_set[4][0] = 8.14
            image_set[4][1] = 1.24
            image_set[4][2] = 6.86
            image_set[4][3] = 2.00
            image_set[4][4] = 5.61
            image_set[4][5] = 2.67
        else:
            image_set[0][0] = 7.20
            image_set[0][1] = 1.16
            image_set[0][2] = 6.46
            image_set[0][3] = 1.77
            image_set[0][4] = 5.85
            image_set[0][5] = 2.00
            image_set[1][0] = 7.67
            image_set[1][1] = 1.24
            image_set[1][2] = 6.70
            image_set[1][3] = 1.69
            image_set[1][4] = 5.49
            image_set[1][5] = 1.63
            image_set[2][0] = 7.22
            image_set[2][1] = 1.40
            image_set[2][2] = 6.31
            image_set[2][3] = 1.95
            image_set[2][4] = 5.39
            image_set[2][5] = 1.66
            image_set[3][0] = 7.46
            image_set[3][1] = 1.28
            image_set[3][2] = 6.62
            image_set[3][3] = 1.83
            image_set[3][4] = 5.80
            image_set[3][5] = 1.92
            image_set[4][0] = 7.67
            image_set[4][1] = 1.97
            image_set[4][2] = 6.02
            image_set[4][3] = 2.50
            image_set[4][4] = 6.49
            image_set[4][5] = 2.31

class Decks(object):
    '''Creates objects that provide deck information for simulation of model
    '''
    def __init__(self, num_decks=4, deck_size=40, set_reward_fun=None, \
                set_punish_fun=None):
        self.reward = [[0] * deck_size for x in range(num_decks)]
        self.punish = [[0] * deck_size for x in range(num_decks)]
        self.num_decks = num_decks
        self.deck_size = deck_size
        self.set_reward_fun = set_reward_fun
        self.set_punish_fun = set_punish_fun
        if (self.set_reward_fun is None):
            self.set_reward_fun = Decks._Becharaetal_2000_rew
        if (self.set_punish_fun is None):
            self.set_punish_fun = Decks._Becharaetal_2000_pun
        self.set_reward_fun(self.reward)
        self.set_punish_fun(self.punish)

    def get_num_decks(self):
        return (self.num_decks)

    def get_deck_size(self):
        return (self.deck_size)

    def get_reward(self):
        return (self.reward)

    def get_punish(self):
        return (self.get_punish)

    @staticmethod
    def _Becharaetal_2000_rew(decks):
        ''' sets the deck to the reward distribution from Bechara et al (2000)

        Arguments
            decks   --  2D list that represents the distribution of rewards for
                        decks
        '''
        for j in range(len(decks[0])):
            decks[0][j] = 100
            decks[1][j] = 100
            decks[2][j] = 50
            decks[3][j] = 50


    @staticmethod
    def _Becharaetal_2000_pun(decks):
        ''' sets the deck to the punishment distribution from Bechara et al (2000)

        Arguments
            decks   --  2D list that represents the distribution of punishments
                        for decks
        '''
        decks[0][2] = 150
        decks[0][4] = 300
        decks[0][6] = 200
        decks[0][8] = 250
        decks[0][9] = 350
        decks[0][11] = 350
        decks[0][13] = 250
        decks[0][14] = 200
        decks[0][16] = 300
        decks[0][17] = 150
        decks[0][21] = 300
        decks[0][23] = 350
        decks[0][25] = 200
        decks[0][26] = 250
        decks[0][27] = 150
        decks[0][30] = 350
        decks[0][31] = 200
        decks[0][32] = 250
        decks[0][36] = 150
        decks[0][37] = 300

        decks[1][8] = 1250
        decks[1][13] = 1250
        decks[1][20] = 1250
        decks[1][31] = 1250

        decks[2][2] = 50
        decks[2][4] = 50;
        decks[2][6] = 50;
        decks[2][8] = 50;
        decks[2][9] = 50;
        decks[2][11] = 25;
        decks[2][12] = 75;
        decks[2][16] = 25;
        decks[2][17] = 75;
        decks[2][19] = 50;
        decks[2][23] = 50;
        decks[2][24] = 25;
        decks[2][25] = 50;
        decks[2][28] = 75;
        decks[2][29] = 50;
        decks[2][33] = 25;
        decks[2][34] = 25;
        decks[2][36] = 75;
        decks[2][38] = 50;
        decks[2][39] = 75;

        decks[3][9] = 250;
        decks[3][19] = 250;
        decks[3][28] = 250;
        decks[3][34] = 250;


